#!/usr/bin/env python

from vogue.settings import templates
from vogue.adapter import VogueAdapter
from vogue.settings import get_vogue_adapter
from fastapi import APIRouter, Depends, Request
from vogue.constants.constants import MONTHS, YEARS, THIS_YEAR
from vogue.crud.find_plots.metric_per_month import value_per_month
from vogue.crud.find_plots.prepps import find_concentration_defrosts, find_concentration_amount
from vogue import __version__

router = APIRouter()


@router.get("/prepps/microbial/")
async def microbial(
    request: Request,
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):
    data = value_per_month(adapter, year, "microbial_library_concentration", "strain")

    return templates.TemplateResponse(
        "microbial.html",
        context=dict(
            request=request,
            header="Microbial Samples",
            data=data,
            months=[m[1] for m in MONTHS],
            version=__version__,
            year_of_interest=year,
            years=YEARS,
        ),
    )


@router.get("/prepps/target_enrichment")
async def target_enrichment(
    request: Request, year: int = THIS_YEAR, adapter: VogueAdapter = Depends(get_vogue_adapter)
):
    library_size_post_hyb = value_per_month(adapter, year, "library_size_post_hyb", "source")
    library_size_pre_hyb = value_per_month(adapter, year, "library_size_pre_hyb", "source")

    return templates.TemplateResponse(
        "target_enrichment.html",
        context=dict(
            request=request,
            header="Target Enrichment (exom/panels)",
            data_pre_hyb=library_size_pre_hyb,
            data_post_hyb=library_size_post_hyb,
            months=[m[1] for m in MONTHS],
            version=__version__,
            year_of_interest=year,
            years=YEARS,
        ),
    )


@router.get("/prepps/wgs")
async def wgs(
    request: Request, year: int = THIS_YEAR, adapter: VogueAdapter = Depends(get_vogue_adapter)
):
    concentration_time = value_per_month(adapter, year, "nr_defrosts-concentration")
    concentration_defrosts = find_concentration_defrosts(adapter=adapter, year=year)

    return templates.TemplateResponse(
        "wgs.html",
        context=dict(
            request=request,
            header="WGS illumina PCR-free",
            concentration_defrosts=concentration_defrosts,
            concentration_time=concentration_time,
            months=[m[1] for m in MONTHS],
            version=__version__,
            year_of_interest=year,
            years=YEARS,
        ),
    )


@router.get("/prepps/lucigen")
async def lucigen(
    request: Request, year: int = THIS_YEAR, adapter: VogueAdapter = Depends(get_vogue_adapter)
):
    amount_concentration_time = value_per_month(adapter, year, "amount-concentration")
    concentration_amount = find_concentration_amount(adapter=adapter, year=year)

    return templates.TemplateResponse(
        "lucigen.html",
        context=dict(
            request=request,
            header="Lucigen PCR-free",
            amount_concentration_time=amount_concentration_time,
            months=[m[1] for m in MONTHS],
            amount=concentration_amount,
            version=__version__,
            year_of_interest=year,
            years=YEARS,
        ),
    )



from vogue.build.reagent_label_category import build_reagent_label_category
import logging

LOG = logging.getLogger(__name__)


def load_all(adapter, lims, categories):
    """Function to load reagent_labels from a step into the database"""

    LOG.info("Fetching all reagent types from lims...")
    lims_reagent_labels = lims.get_reagent_types()
    LOG.info("Loading reagent types into the reagent_label_cathegory colleciton.")
    for lims_reagent_label in lims_reagent_labels:
        if lims_reagent_label.category not in categories:
            continue
        mongo_reagent_label = build_reagent_label_category(lims_reagent_label)
        adapter.add_or_update_document(
            mongo_reagent_label, adapter.reagent_label_category_collection
        )



from typing import Optional


from vogue.adapter import VogueAdapter
from vogue.models.database.sample import Sample
from vogue.models.database.bioinfo_sample.base import BioInfoSample


def find_sample(adapter: VogueAdapter, sample_id: str) -> Optional[Sample]:
    """Find one sample from the sample collection"""

    db_sample = adapter.sample_collection.find_one({"_id": sample_id})
    if not db_sample:
        return None

    return Sample(**db_sample)


def find_bioinfo_sample(adapter: VogueAdapter, sample_id: str) -> Optional[BioInfoSample]:
    """Find one sample from the sample collection"""

    db_sample = adapter.bioinfo_samples_collection.find_one({"_id": sample_id})
    if not db_sample:
        return None

    return BioInfoSample(**db_sample)



#!/usr/bin/env python

from vogue.constants.constants import MONTHS, DNA_PICARD

VALID_WORKFLOWS = ["balsamic", "mip-dna"]
VALID_APPLICATION = ["wgs", "tga", "wes"]


def qc_dna_picard_time_plot(adapter, year: int) -> dict:
    """Prepares data for the picard over time plot."""

    pipe = [
        {
            "$lookup": {
                "from": "sample",
                "localField": "_id",
                "foreignField": "_id",
                "as": "sample_info",
            }
        },
        {"$unwind": {"path": "$sample_info"}},
        {"$match": {"sample_info.received_date": {"$exists": "True"}}},
        {
            "$project": {
                "mip-dna": 1,
                "balsamic": 1,
                "workflows": 1,
                "received_date": "$sample_info.received_date",
                "category": "$sample_info.category",
            }
        },
        {"$match": {"received_date": {"$exists": "True"}}},
        {
            "$project": {
                "month": {"$month": "$received_date"},
                "year": {"$year": "$received_date"},
                "mip-dna": 1,
                "balsamic": 1,
                "workflows": 1,
                "category": 1,
            }
        },
        {"$match": {"year": {"$eq": int(year)}}},
    ]
    aggregate_result = adapter.bioinfo_samples_aggregate(pipe)
    final_data = {}
    for data in DNA_PICARD.values():
        for key in data:
            final_data[key] = {}

    for sample in aggregate_result:
        qc_dna_analysis = sample.get("balsamic")
        if not qc_dna_analysis:
            qc_dna_analysis = sample.get("mip-dna")
            if not qc_dna_analysis:
                continue
        qc_data_to_process = qc_dna_analysis.get("multiqc_picard_insertSize")
        if qc_data_to_process:
            final_data = _append_to_final_data(
                final_data=final_data, raw_data=qc_data_to_process, sample=sample
            )

        qc_data_to_process = qc_dna_analysis.get("multiqc_picard_HsMetrics")
        if qc_data_to_process:
            final_data = _append_to_final_data(
                final_data=final_data, raw_data=qc_data_to_process, sample=sample
            )

    plot_data = {"data": final_data, "labels": [m[1] for m in MONTHS]}
    return plot_data


def _append_to_final_data(final_data: dict, raw_data: dict, sample: dict):
    """Parsing raw data. Appending to final_data"""

    for key, val in raw_data.items():
        if key in final_data.keys():
            category = f"{sample['category']}_{'_'.join(sample['workflows'])}"
            if category not in final_data[key].keys():
                final_data[key][category] = []
            final_data[key][category].append(
                {
                    "name": sample["_id"],
                    "application": sample["category"],
                    "workflows": ",".join(sample["workflows"]),
                    "x": sample["month"],
                    "y": val,
                }
            )
    return final_data


def qc_dna_picard_plot(adapter, year: int) -> dict:
    """Prepares data for the picard plot."""

    all_samples = adapter.bioinfo_samples_collection.find()
    final_data = []

    for sample in all_samples:
        sample_id = sample["_id"]
        qc_dna_analysis = sample.get("balsamic")
        if not qc_dna_analysis:
            qc_dna_analysis = sample.get("mip-dna")
            if not qc_dna_analysis:
                continue
        multiqc_picard_insertSize = qc_dna_analysis.get("multiqc_picard_insertSize")
        multiqc_picard_HsMetrics = qc_dna_analysis.get("multiqc_picard_HsMetrics")
        merged = multiqc_picard_insertSize.copy()
        merged.update(multiqc_picard_HsMetrics)
        merged["_id"] = sample_id
        final_data.append(merged)

    plot_data = {"final_data": final_data, "labels": [m[1] for m in MONTHS]}

    return plot_data



#!/usr/bin/env python
import logging
import json
from dateutil.parser import parse

LOG = logging.getLogger(__name__)


def load_sample(adapter, genotype_sample_string):

    LOG.info("Trying to build json")
    try:
        mongo_sample = json.loads(genotype_sample_string)
        if mongo_sample.get("sample_created_in_genotype_db"):
            date = parse(mongo_sample["sample_created_in_genotype_db"])
            mongo_sample["sample_created_in_genotype_db"] = date
    except ValueError as e:
        LOG.error(e)
        return

    if "_id" not in mongo_sample:
        LOG.error("Not a propper mongo document. Missing _id")
        return

    adapter.add_or_update_document(mongo_sample, adapter.genotype_analysis_collection)



import logging
import click
import json
from vogue.load.application_tag import load_aplication_tags

LOG = logging.getLogger(__name__)


@click.command("apptag", short_help="Reads json with application tag info.")
@click.argument("application-tags")
@click.pass_context
def application_tags(ctx: click.Context, application_tags):
    """Reads json string with application tags. Eg:'[{"tag":"MELPCFR030", "category":"wgs",...},...]'"""
    adapter = ctx.obj["adapter"]
    LOG.info("Reading json.")
    try:
        json_list = json.loads(application_tags)
    except json.JSONDecodeError as e:
        LOG.error("Imput string is not proper json.")
        LOG.error(e)
        raise click.Abort()

    if not isinstance(json_list, list):
        LOG.error("Cannot read input. Its not list of jsons.")
        raise click.Abort()

    LOG.info("json is read.")
    load_aplication_tags(adapter, json_list)



import json
import yaml
import logging
import os
import pathlib
import copy
import collections

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
LOG = logging.getLogger(__name__)


def convert_dot(string):
    """
    replaces dot with underscore
    """
    return string.replace(".", "_")


def add_doc(docstring):
    """
    A decorator for adding docstring. Taken shamelessly from stackexchange.
    """

    def document(func):
        func.__doc__ = docstring
        return func

    return document


def dict_replace_dot(obj):
    """
    recursively replace all dots in json.load keys.
    """
    if isinstance(obj, dict):
        for key in obj.keys():
            obj[convert_dot(key)] = obj.pop(key)
            if isinstance(obj[convert_dot(key)], dict) or isinstance(obj[convert_dot(key)], list):
                obj[convert_dot(key)] = dict_replace_dot(obj[convert_dot(key)])
    elif isinstance(obj, list):
        for item in obj:
            item = dict_replace_dot(item)
    return obj


def json_read(fname):
    """
    Reads JSON file and returns dictionary. Returns error if can't read.
    """

    try:
        with open(fname, "r") as f:
            analysis_conf = json.load(f, object_hook=dict_replace_dot)
            return analysis_conf
    except:
        LOG.warning("Input config is not JSON")
        return False


def yaml_read(fname):
    """
    Reads YAML file and returns dictionary. Returns error if can't read.
    """

    try:
        with open(fname, "r") as f:
            analysis_conf = yaml.load(f)
            return analysis_conf
    except:
        LOG.warning("Input config is not YAML")
        return False


def check_file(fname):
    """
    Check file exists and readable.
    """

    path = pathlib.Path(fname)

    if not path.exists() or not path.is_file():
        LOG.error("File not found or input is not a file.")
        raise FileNotFoundError


def concat_dict_keys(my_dict: dict, key_name="", out_key_list=list()):
    """
    Recursively create a list of key:key1,key2 from a nested dictionary
    """

    if isinstance(my_dict, dict):

        if key_name != "":
            out_key_list.append(key_name + ":" + ", ".join(list(my_dict.keys())))

        for k in my_dict.keys():
            concat_dict_keys(my_dict[k], key_name=k, out_key_list=out_key_list)

    return out_key_list


def recursive_default_dict():
    """
    Recursivly create defaultdict.
    """
    return collections.defaultdict(recursive_default_dict)


def convert_defaultdict_to_regular_dict(inputdict: dict):
    """
    Recursively convert defaultdict to dict.
    """
    if isinstance(inputdict, collections.defaultdict):
        inputdict = {
            key: convert_defaultdict_to_regular_dict(value) for key, value in inputdict.items()
        }
    return inputdict



from datetime import date

MONTHS = [
    (1, "Jan"),
    (2, "Feb"),
    (3, "Mar"),
    (4, "Apr"),
    (5, "May"),
    (6, "Jun"),
    (7, "Jul"),
    (8, "Aug"),
    (9, "Sep"),
    (10, "Oct"),
    (11, "Nov"),
    (12, "Dec"),
]

THIS_YEAR: int = date.today().year

TEST_SAMPLES = [
    "SIB603A6",
    "ADM1852A2",
    "SIB603A3",
    "SER305A1",
    "AND51A3",
    "BRA1907A5",
    "SIB552A11",
    "SIB603A9",
    "SIB257A27",
    "SIB552A15",
    "SIB611A3",
    "SIB552A16",
    "SIB256A3",
    "SIB601A5",
    "SIB553A24",
    "SIB404A15",
    "SIB404A13",
    "SIB663A2",
    "SIB701A1",
    "SIB552A18",
    "SIB257A26",
    "SER303A1",
    "SIB553A27",
    "SIB257A23",
    "SIB607A1",
    "SIB603A2",
    "SIB404A14",
    "SIB403A27",
    "SIB603A12",
    "SIB603A14",
    "SIB601A11",
    "SER305A2",
    "SIB663A4",
    "ADM1852A1",
    "SIB552A12",
    "SIB552A13",
    "SIB552A10",
    "SIB553A19",
    "SIB553A10",
    "SIB553A15",
    "SIB603A8",
    "SIB603A4",
    "SIB553A17",
    "SIB603A16",
    "SIB652A1",
    "SIB552A3",
    "SIB552A4",
    "SIB601A7",
    "SIB553A28",
    "SIB552A2",
    "SIB603A18",
    "BRA1907A1",
    "SIB404A17",
    "SIB404A16",
    "SIB553A3",
    "SIB256A1",
    "BIL1393A1",
    "SIB601A1",
    "SIB603A10",
    "SER305A5",
    "SIB603A5",
    "SER305A4",
    "ADM1851A4",
    "SER305A6",
    "SER307A8",
    "BRA1907A4",
    "BRA1907A3",
    "SIB601A4",
    "SER307A7",
    "BRA1907A2",
    "SIB257A25",
    "SIB603A13",
    "SIB601A12",
    "SIB601A14",
    "SER305A7",
    "SIB552A17",
    "SIB256A2",
    "SIB553A14",
    "SIB601A2",
    "SIB553A12",
    "SIB603A20",
    "SIB553A13",
    "SIB553A9",
    "SIB404A19",
    "ADM1852A3",
    "SIB606A2",
    "SIB404A18",
    "SIB552A5",
    "SIB603A1",
    "SIB552A6",
    "SER305A3",
    "SIB552A8",
    "SIB351A22",
    "SER307A4",
    "SIB603A15",
    "SER305A8",
    "SIB603A19",
    "SIB553A7",
    "SIB611A1",
    "SIB553A8",
    "SIB553A5",
    "SER303A3",
    "SIB351A21",
    "SIB553A16",
    "SIB255A1",
    "SIB608A1",
    "SIB351A24",
    "SIB501A2",
    "SIB403A25",
    "SIB501A3",
    "SER303A7",
    "SIB603A11",
    "AND51A4",
    "SIB553A25",
    "SIB502A2",
    "SIB501A4",
    "SER303A8",
    "SIB652A2",
    "SIB257A24",
    "AND51A1",
    "BRA1907A6",
    "AND51A5",
    "SIB501A1",
    "SIB601A8",
    "SIB257A22",
    "SIB653A1",
    "SER307A2",
    "SER303A6",
    "SIB351A20",
    "SER307A3",
    "SIB553A6",
    "ADM1852A4",
    "SIB601A6",
    "SIB404A21",
    "SER303A4",
    "SER303A5",
    "SER303A2",
    "SIB601A10",
    "SIB553A26",
    "SER307A5",
    "SIB553A1",
    "SIB253A1",
    "SIB351A23",
    "SIB253A2",
    "SIB404A22",
    "SIB601A3",
    "SIB553A4",
    "SIB702A1",
    "ADM1851A3",
    "SIB552A14",
    "AND51A6",
    "SIB609A1",
    "SER307A6",
    "SIB606A1",
    "SIB611A2",
    "SIB553A22",
    "SIB255A3",
    "SIB603A17",
    "SIB253A3",
    "ADM1851A2",
    "SIB601A13",
    "SIB553A18",
    "AND51A2",
    "SIB403A26",
    "SIB552A9",
    "SIB553A2",
    "SIB553A20",
    "SER307A1",
    "SIB553A21",
    "SIB404A20",
    "SIB663A3",
    "SIB404A24",
    "ADM990A1",
    "SIB603A7",
    "SIB404A23",
    "SIB553A23",
    "SIB552A7",
    "SIB502A1",
    "SIB351A19",
    "SIB552A1",
    "SIB601A9",
    "ADM1851A1",
    "SIB553A11",
    "SIB255A2",
]

YEARS = [y for y in range(2016, THIS_YEAR + 1)]

LANE_UDFS = [
    "Clusters PF R1",
    "Clusters PF R2",
    "% Aligned R1",
    "% Aligned R2",
    "% Bases >=Q30",
    "% Bases >=Q30 R1",
    "% Bases >=Q30 R2",
    "% Error Rate R1",
    "% Error Rate R2",
    "% Phasing R1",
    "% Prephasing R1",
    "% Prephasing R2",
    "%PF R1",
    "%PF R2",
    "Cluster Density (K/mm^2) R1",
    "Cluster Density (K/mm^2) R2",
    "Intensity Cycle 1 R1",
    "Intensity Cycle 1 R2",
    "Reads PF (M) R1",
    "Reads PF (M) R2",
    "Yield PF (Gb) R1",
    "Yield PF (Gb) R2",
    "% Phasing R2",
]

RUN_TYPES = [
    "AUTOMATED - NovaSeq Run",
    "CG002 - Illumina Sequencing (Illumina SBS)",
    "CG002 - Illumina Sequencing (HiSeq X)",
]

BIOINFO_HELP_URLS = {
    "PICARD_INSERT_SIZE": {
        "url": "https://broadinstitute.github.io/picard/picard-metric-definitions.html#InsertSizeMetrics",
        "text": "Metrics about the insert size distribution of a paired-end library, created by the CollectInsertSizeMetrics program",
    },
    "PICARD_HS_METRIC": {
        "url": "https://broadinstitute.github.io/picard/picard-metric-definitions.html#HsMetrics",
        "text": "Metrics generated by CollectHsMetrics for the analysis of target-capture sequencing experiments.",
    },
}

DNA_PICARD = {
    "PICARD_INSERT_SIZE": [
        "MEDIAN_INSERT_SIZE",
        "MEDIAN_ABSOLUTE_DEVIATION",
        "MIN_INSERT_SIZE",
        "MAX_INSERT_SIZE",
        "MEAN_INSERT_SIZE",
        "STANDARD_DEVIATION",
        "READ_PAIRS",
    ],
    "PICARD_HS_METRIC": [
        "PCT_EXC_DUPE",
        "PCT_EXC_MAPQ",
        "PCT_EXC_BASEQ",
        "PCT_EXC_OVERLAP",
        "PCT_EXC_OFF_TARGET",
        "FOLD_80_BASE_PENALTY",
        "PCT_TARGET_BASES_1X",
        "PCT_TARGET_BASES_2X",
        "PCT_TARGET_BASES_10X",
        "PCT_TARGET_BASES_20X",
        "PCT_TARGET_BASES_30X",
        "PCT_TARGET_BASES_40X",
        "PCT_TARGET_BASES_50X",
        "PCT_TARGET_BASES_100X",
        "PCT_TARGET_BASES_250X",
        "PCT_TARGET_BASES_1000X",
        "HS_LIBRARY_SIZE",
        "HS_PENALTY_10X",
        "HS_PENALTY_20X",
        "HS_PENALTY_30X",
        "HS_PENALTY_40X",
        "HS_PENALTY_50X",
        "HS_PENALTY_100X",
        "AT_DROPOUT",
        "GC_DROPOUT",
        "HET_SNP_SENSITIVITY",
        "HET_SNP_Q",
        "TARGET_TERRITORY",
        "BAIT_DESIGN_EFFICIENCY",
        "TOTAL_READS",
        "PF_READS",
        "PF_UNIQUE_READS",
        "PCT_PF_READS",
        "PCT_PF_UQ_READS",
        "PF_UQ_READS_ALIGNED",
        "PCT_PF_UQ_READS_ALIGNED",
        "PF_BASES_ALIGNED",
        "PF_UQ_BASES_ALIGNED",
        "ON_BAIT_BASES",
        "NEAR_BAIT_BASES",
        "OFF_BAIT_BASES",
        "ON_TARGET_BASES",
        "PCT_SELECTED_BASES",
        "PCT_OFF_BAIT",
        "ON_BAIT_VS_SELECTED",
        "MEAN_BAIT_COVERAGE",
        "MEAN_TARGET_COVERAGE",
        "MEDIAN_TARGET_COVERAGE",
        "MAX_TARGET_COVERAGE",
        "PCT_USABLE_BASES_ON_BAIT",
        "PCT_USABLE_BASES_ON_TARGET",
        "FOLD_ENRICHMENT",
        "ZERO_CVG_TARGETS_PCT",
        "GENOME_SIZE",
        "BAIT_TERRITORY",
    ],
}

INSTRUMENTS = {
    "D00410": "Sigourney",
    "D00415": "Merida",
    "D00450": "Arnold",
    "ST-E00198": "Dorothy",
    "ST-E00201": "Irene",
    "ST-E00214": "Marie",
    "ST-E00266": "Rita",
    "ST-E00269": "Rosalyn",
    "A00187": "Ingrid",
    "A00689": "Barbara",
    "A00621": "Greta",
    "FS10000534": "Katherine",
}

MICROSALT = {
    "picard_markduplicate": {
        "insert_size": "Median distance between read 1 and read 2 of paired reads. Trimmed deduplicated reads mapped via bwa, reported in Picard. Result is extrapolated from curve fitting the data.",
        "duplication_rate": "Percentage of reads that were exact duplicates (thus giving no new information). Trimmed reads mapped via bwa, reported in picard",
    },
    "microsalt_samtools_stats": {
        "average_coverage": "Summed coverage of each position, divided by reference genome length. Trimmed deduplicated reads mapped via bwa, reported in  Samtools",
        "coverage_10x": "Number of positions above 10x coverage, divided by genome length	Trimmed deduplicated reads mapped via bwa, reported in  Samtools",
        "coverage_30x": "Number of positions above 30x coverage, divided by genome length	Trimmed deduplicated reads mapped via bwa, reported in  Samtools",
        "coverage_50x": "Number of positions above 50x coverage, divided by genome length	Trimmed deduplicated reads mapped via bwa, reported in  Samtools",
        "coverage_100x": "Number of positions above 100x coverage, divided by genome length	Trimmed deduplicated reads mapped via bwa, reported in  Samtools",
        "total_reads": "Number of reads	Trimmed reads mapped via bwa, reported in samtools.",
        "mapped_rate": "Percentage of reads that mapped to the reference. Trimmed reads mapped via bwa, reported in samtools.",
    },
    "quast_assembly": {"necessary_contigs", "gc_percentage", "estimated_genome_length", "n50"},
}



from typing import Literal, List
from pydantic import BaseModel


class BlastPubmlst(BaseModel):
    sequence_type: int
    thresholds: Literal["Passed", "Failed"]


class QuastAssembly(BaseModel):
    estimated_genome_length: int
    gc_percentage: float
    n50: int
    necessary_contigs: int


class PicardMarkDuplicate(BaseModel):
    insert_size: int
    duplication_rate: float


class MicroSaltSamtoolsStats(BaseModel):
    total_reads: int
    mapped_rate: float
    average_coverage: float
    coverage_10x: float
    coverage_30x: float
    coverage_50x: float
    coverage_100x: float


class MicroSalt(BaseModel):
    blast_pubmlst: BlastPubmlst
    quast_assembly: QuastAssembly
    blast_resfinder_resistence: List[str]
    picard_markduplicate: PicardMarkDuplicate
    microsalt_samtools_stats: MicroSaltSamtoolsStats



from genologics.entities import Process

from vogue.parse.build.reagent_label import reagent_label_data


def build_reagent_label(step: Process) -> dict:
    """Build reagent label document from lims data."""
    reagent_labels = reagent_label_data(step)

    return reagent_labels



from vogue.build.reagent_label import build_reagent_label
from vogue.constants.lims_constants import MASTER_STEPS_UDFS, INSTRUMENTS
import logging

LOG = logging.getLogger(__name__)

REAGENT_LABEL_STEPS = MASTER_STEPS_UDFS["reagent_labels"]["steps"]["bcl"]


def load_one(adapter, step):
    """Function to load reagent_labels from a step into the database"""
    LOG.info("Processing reagent_labels from step: %s" % step.id)
    mongo_reagent_labels = build_reagent_label(step)
    for reagent_label, mongo_reagent_label in mongo_reagent_labels.items():
        adapter.add_or_update_document(mongo_reagent_label, adapter.reagent_label_collection)


def load_all(adapter, lims):
    """Function to load reagent_labels from all lims flowcells into the database"""
    processes = lims.get_processes(type=REAGENT_LABEL_STEPS)
    LOG.info("Loading data from %s processes" % str(len(processes)))
    for step in processes:
        load_one(adapter, step)


def load_recent(adapter, lims, the_date):
    """Function to load reagent_labels from all lims flowcells run after the_date into the database"""
    processes = lims.get_processes(type=REAGENT_LABEL_STEPS, last_modified=the_date)
    LOG.info("Loading data from %s processes" % str(len(processes)))
    for step in processes:
        load_one(adapter, step)



ANALYSIS_DESC = {
    "multiqc": "General quality control metrics for post- and pre-variant calling prepared by multiqc",
    "microsalt": "Results and quality control metrics output from microsalt workflow",
}

ANALYSIS_SETS = {
    "multiqc": {
        "multiqc_picard_dups": "multiqc json/yaml report key for picard mark duplicate.",
        "multiqc_picard_HsMetrics": "multiqc json/yaml report key for picard collecthsmetrics.",
        "multiqc_picard_AlignmentSummaryMetrics": "multiqc json/yaml report key for picard alignment summary metrics",
        "multiqc_picard_insertSize": "multiqc json/yaml report key for picard insert size metrics",
    },
    "microsalt": {
        "blast_pubmlst": "results for blasting for molecular typing and microbial genome diversity against pubmslt database",
        "quast_assembly": "results for quality assessment tool for genome assemblies (quast)",
        "blast_resfinder_resistence": "results blasting against the resistance database provided by resfinder db",
        "picard_markduplicate": "quality control and marking duplicate reads in bam/sam files",
        "microsalt_samtools_stats": "quality control to collect statistics from bam files",
    },
}



from pydantic import BaseModel


class MultiqcPicardAlignmentSummaryMetrics(BaseModel):
    total_reads: int  # ?
    pct_pf_reads_aligned: float
    mean_read_length: float


class MultiqcPicardInsertSize(BaseModel):
    medin_insert_size: int


class MultiqcPicardWGSMetrics(BaseModel):
    median_coverage: float
    pct_10x: float
    pct_30x: float
    pct_50x: float
    pct_100x: float


class MultiqcCutadapt(BaseModel):
    percent_trimmed: float


class MultiqcGeneralStats(BaseModel):
    Fastqc_mqc_generalstats_fastqc_percent_duplicates: float
    Fastqc_mqc_generalstats_fastqc_percent_gc: float
    Fastqc_mqc_generalstats_fastqc_total_sequences: float


class Mutant(BaseModel):
    multiqc_picard_alignment_summary_metrics: MultiqcPicardAlignmentSummaryMetrics
    multiqc_picard_insertSize: MultiqcPicardInsertSize
    multiqc_picard_wgsmetrics: MultiqcPicardWGSMetrics
    multiqc_cutadapt: MultiqcCutadapt
    multiqc_general_stats: MultiqcGeneralStats



#!/usr/bin/env python


def get_genotype_plate(adapter, plate_id: str) -> dict:
    plates_pipe = [
        {"$match": {"plate": {"$exists": "True"}, "snps.comp": {"$exists": "True"}}},
        {"$group": {"_id": {"plate": "$plate"}}},
    ]
    aggregate_result = list(adapter.genotype_analysis_aggregate(plates_pipe))
    plates = [plate["_id"]["plate"] for plate in aggregate_result]

    plate_id = plates[0] if not plate_id else plate_id

    samples_pipe = [{"$match": {"plate": plate_id, "snps.comp": {"$exists": "True"}}}]

    samples = list(adapter.genotype_analysis_aggregate(samples_pipe))
    data = []
    x_labels = list(samples[0]["snps"]["comp"].keys())
    y_labels = []
    x_labels.sort()
    for row, sample in enumerate(samples):
        comp = sample["snps"].get("comp")
        genotype = sample["snps"].get("genotype")
        sequence = sample["snps"].get("sequence")
        y_labels.append(sample["_id"])
        for col, key in enumerate(x_labels):
            internal = "".join(genotype[key])
            external = "".join(sequence[key])
            data.append((col, row, int(comp[key]), f"{internal} : {external}"))

    return {
        "data": data,
        "x_labels": x_labels,
        "y_labels": y_labels,
        "plates": plates,
        "plate_id": plate_id,
    }



"""
Module with CLI commands for vogue
The CLI is intended for development/testing purpose only. To run in a production setting please refer to documentation
for suggestions how.
"""

import logging
import click
import uvicorn

# Get version and doc decorator
from vogue import __version__
from vogue.settings import settings
from .load import load

LOG = logging.getLogger(__name__)


@click.version_option(__version__)
@click.group()
@click.pass_context
def cli(context: click.Context):
    """Main entry point"""
    logging.basicConfig(level=logging.INFO)
    context.obj = {}


@cli.command(name="serve")
@click.option("--version", default="v1", type=click.Choice(["v1", "v2"]), show_default=True)
@click.option("--reload", is_flag=True)
def serve_command(reload: bool, version: str):
    """Serve the vogue app for testing purpose."""

    app = f"vogue.api.api_{version}.api:app"
    LOG.info("Running on host:%s and port:%s", settings.host, settings.port)
    uvicorn.run(app=app, host=settings.host, port=settings.port, reload=reload)


cli.add_command(load)



from typing import Optional
from fastapi import status


class VogueError(Exception):
    def __init__(self, message: str):
        self.message = message


class MissingApplicationTag(Exception):
    pass


class VogueRestError(VogueError):
    def __init__(self, message: str, code: Optional[int] = None):
        self.message = message
        self.code = code
        super().__init__(message)


class InsertError(VogueRestError):
    def __init__(self, message: str, code: Optional[int] = status.HTTP_405_METHOD_NOT_ALLOWED):
        self.message = message
        self.code = code
        super().__init__(message, code)



#!/usr/bin/env python
from typing import Optional

from vogue.settings import templates

from vogue.adapter import VogueAdapter
from vogue.settings import get_vogue_adapter
from fastapi import APIRouter, Depends, Request

router = APIRouter()

from vogue.constants.constants import YEARS, MICROSALT, THIS_YEAR
from vogue.crud.find_plots.bioinfo.micro import (
    microsalt_get_untyped,
    microsalt_get_st_time,
    microsalt_get_qc_time,
    microsalt_get_strain_st,
)
from vogue import __version__


@router.get("/Bioinfo/Microbial/strain_st")
async def micro_strain_st(
    request: Request,
    strain: Optional[str] = "",
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):
    results = microsalt_get_strain_st(adapter, year)
    if results and not strain:
        strain = list(results.keys())[0]
    return templates.TemplateResponse(
        "microsalt_strain_st.html",
        context=dict(
            request=request,
            data=results.get(strain, {}),
            strain=strain,
            categories=results.keys(),
            header="Microsalt",
            version=__version__,
            year_of_interest=year,
            years=YEARS,
        ),
    )


@router.get("/Bioinfo/Microbial/qc_time")
async def micro_qc_time(
    request: Request,
    metric_path: str = "picard_markduplicate.insert_size",
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):
    results = microsalt_get_qc_time(adapter, year=year, metric_path=metric_path, category="mic")

    return templates.TemplateResponse(
        "microsalt_qc_time.html",
        context=dict(
            request=request,
            results=results["data"],
            outliers=results["outliers"],
            categories=results["labels"],
            mean=results["mean"],
            selected_group=metric_path.split(".")[0],
            selected_metric=metric_path.split(".")[1],
            header="Microsalt",
            version=__version__,
            year_of_interest=year,
            MICROSALT=MICROSALT,
            years=YEARS,
        ),
    )


@router.get("/Bioinfo/Microbial/untyped")
def micro_untyped(
    request: Request,
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):
    results = microsalt_get_untyped(adapter, year)

    return templates.TemplateResponse(
        "microsalt_untyped.html",
        context=dict(
            request=request,
            results=results["data"],
            categories=results["labels"],
            header="Microsalt",
            version=__version__,
            year_of_interest=year,
            MICROSALT=MICROSALT,
            years=YEARS,
        ),
    )


@router.get(
    "/Bioinfo/Microbial/st_time",
)
async def micro_st_time(
    request: Request,
    strain: str = "",
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):
    results_all = microsalt_get_st_time(adapter, year)
    if results_all["data"] and not strain:
        strain = list(results_all["data"].keys())[0]
    strain_results = results_all["data"].get(strain, {})

    return templates.TemplateResponse(
        "microsalt_st_time.html",
        context=dict(
            request=request,
            results=strain_results,
            results_sorted_keys=sorted(strain_results.keys()),
            strain=strain,
            strains=results_all["data"].keys(),
            categories=results_all["labels"],
            header="Microsalt",
            version=__version__,
            year_of_interest=year,
            MICROSALT=MICROSALT,
            years=YEARS,
        ),
    )



#!/usr/bin/env python


from vogue.settings import templates

from vogue.adapter import VogueAdapter
from vogue.settings import get_vogue_adapter
from fastapi import APIRouter, Depends, Request

router = APIRouter()
from vogue.constants.constants import MICROSALT, YEARS, THIS_YEAR
from vogue.crud.find_plots.bioinfo.micro import microsalt_get_qc_time
from vogue.crud.find_plots.bioinfo.covid import get_qc
from vogue import __version__

HEADER = "Microsalt Covid"


@router.get(
    "/Bioinfo/Covid/qc_time",
)
async def covid_qc_time(
    request: Request,
    metric_path: str = "picard_markduplicate.insert_size",
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):
    """Box plot with qc data per month"""

    results = microsalt_get_qc_time(adapter, year=year, metric_path=metric_path, category="cov")
    return templates.TemplateResponse(
        "microsalt_qc_time.html",
        context=dict(
            request=request,
            results=results["data"],
            outliers=results["outliers"],
            categories=results["labels"],
            mean=results["mean"],
            selected_group=metric_path.split(".")[0],
            selected_metric=metric_path.split(".")[1],
            header=HEADER,
            version=__version__,
            year_of_interest=year,
            MICROSALT=MICROSALT,
            years=YEARS,
        ),
    )


@router.get(
    "/Bioinfo/Covid/qc_time_scatter",
)
async def covid_qc_scatter(
    request: Request,
    metric_path: str = "picard_markduplicate.insert_size",
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):
    """Scatter plot with qc data over time, grouped by prep method"""

    results = get_qc(adapter, year=year, metric_path=metric_path)

    return templates.TemplateResponse(
        "cov_qc_scatter.html",
        context=dict(
            request=request,
            results=results,
            MICROSALT=MICROSALT,
            selected_group=metric_path.split(".")[0],
            metric=metric_path.split(".")[1],
            header=HEADER,
            version=__version__,
            year_of_interest=year,
            years=YEARS,
        ),
    )



import logging
import click
from vogue.load.genotype import load_sample

LOG = logging.getLogger(__name__)


@click.command("genotype", short_help="Load genotype document into db.")
@click.option("-s", "--sample-doc", help="Dictionary with genotype data.")
@click.pass_context
def genotype(ctx: click.Context, sample_doc):
    """Reads data from the genotype database collection and parses out plot specific data on sample
    level to put into the genotype_analysis colleciton"""
    adapter = ctx.obj["adapter"]

    if sample_doc:
        load_sample(adapter, sample_doc)
        return



import sys
import logging
from vogue.build.application_tag import build_application_tag
from vogue.exceptions import MissingApplicationTag

LOG = logging.getLogger(__name__)


def load_aplication_tags(adapter, json_list):
    """Will go through all application tags in json_list and add/update them to trending-db.

    Args:
        adapter(adapter.VogueAdapter)
        json_list(list(dict)): [{'tag':'MELPCFR030', 'category':'wgs',...},...]
    """

    for application_tag in json_list:
        try:
            mongo_application_tag = build_application_tag(application_tag)
            adapter.add_or_update_document(mongo_application_tag, adapter.app_tag_collection)
        except MissingApplicationTag:
            LOG.warning("ApplicationTag missing in JSON list")



"""
    Functionality to add or update to processed bioinfo collection
"""
import logging
import copy
import click


from vogue.tools.cli_utils import concat_dict_keys
from vogue.tools.cli_utils import add_doc as doc
from vogue.tools.cli_utils import recursive_default_dict
from vogue.tools.cli_utils import convert_defaultdict_to_regular_dict
from vogue.build.bioinfo_analysis import build_analysis
from vogue.load.bioinfo_analysis import load_analysis
from vogue.parse.load.bioinfo_analysis import inspect_analysis_result
import vogue.models.bioinfo_analysis as analysis_model

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
LOG = logging.getLogger(__name__)


@click.command(
    "process",
    short_help="""Process stats and result from bioinfo raw and load into the bioinfo_processed collection. """,
)
@click.option(
    "-t",
    "--analysis-type",
    type=click.Choice(list(analysis_model.ANALYSIS_DESC.keys()) + ["all"]),
    multiple=True,
    default=["all"],
    help="Type of analysis results to load.",
)
@click.option(
    "-c",
    "--analysis-case",
    help="""The case that this sample belongs.
        It can be specified multiple times.""",
)
@click.option(
    "-w",
    "--analysis-workflow",
    type=click.Choice(["mip-dna", "balsamic", "microsalt"]),
    required=True,
    help="Analysis workflow used.",
)
@click.option("--workflow-version", required=True, help="Analysis workflow used.")
@click.option(
    "--cleanup/--not-cleanup",
    is_flag=True,
    help="Specify this flag if input json should be cleanup based on analysis-type and models.",
)
@click.option(
    "--case-analysis-type",
    type=click.Choice(["multiqc", "microsalt", "custom"]),
    default="multiqc",
    help="Specify the type for the case analysis. i.e. if it is multiqc output, then choose multiqc",
)
@click.option(
    "-a",
    "--load-all",
    is_flag=True,
    help="Load all samples for all cases within bioinfo_processed.",
)
@click.option("--dry", is_flag=True, help="Load from sample or not. (dry-run)")
@doc(
    f"""
    Load bioinfo analysis results from bioinfo_raw collection into bioinfo_processed 

    Analysis types recognize the following keys in the input file: {" ".join(concat_dict_keys(analysis_model.ANALYSIS_SETS,key_name=""))}
        """
)
@click.pass_context
def bioinfo_process(
    ctx: click.Context,
    dry,
    analysis_type,
    analysis_case,
    analysis_workflow,
    workflow_version,
    case_analysis_type,
    cleanup,
    load_all,
):
    adapter = ctx.obj["adapter"]

    if (not load_all and not analysis_case) or (load_all and analysis_case):
        LOG.error(
            "--load-all and --analysis-case are mutually exclusive and cannot be used together"
        )
        raise click.Abort()
    analysis_dict = dict()

    # if is_case flag is enabled, build dictionary without merging.
    # Loop over list of input config files for single sample and merge them into
    # one single dictionary

    valid_analysis = dict()
    analysis_dict["case"] = analysis_case
    analysis_dict["workflow"] = analysis_workflow
    analysis_dict["workflow_version"] = workflow_version
    analysis_dict["case_analysis_type"] = case_analysis_type

    current_analysis = adapter.bioinfo_raw(analysis_case)

    if current_analysis is None:
        LOG.info("Raw import for this case does not exist. Load it without processed flag first")
        raise click.Abort()

    if not case_analysis_type in current_analysis["case_analysis_types"]:
        LOG.info("%s doesn't exist for case %s", case_analysis_type, analysis_case)
        raise click.Abort()

    analysis_dict[case_analysis_type] = current_analysis[analysis_workflow][case_analysis_type][-1]

    analysis_dict["sample"] = current_analysis["samples"]
    # if case analysis type is microsalt, aggregate samples under analysis result keys
    # e.g. {'smpl_1': {'key':'value_smpl1'}, 'smpl_2': {'key':'value_smpl2'}}
    # will be converted to {'key': {'smpl_1':'value_smpl1', 'smpl_2': 'value_smpl2'}}
    if case_analysis_type == "microsalt":
        new_analysis_dict = recursive_default_dict()
        for key in analysis_dict["microsalt"].keys():
            if key in analysis_dict["sample"]:
                for next_key, next_value in analysis_dict["microsalt"][key].items():
                    new_analysis_dict[next_key] = {
                        **new_analysis_dict[next_key],
                        **{key: next_value},
                    }
        analysis_dict.pop("microsalt")
        tmp_dict = convert_defaultdict_to_regular_dict(new_analysis_dict)
        analysis_dict["microsalt"] = copy.deepcopy(tmp_dict)

    if cleanup:
        LOG.info("Validating parsed config file(s).")
        valid_analysis = inspect_analysis_result(analysis_dict)
        if valid_analysis is None:
            LOG.error("Invalid or badly formatted file(s).")
            raise click.Abort()

    current_analysis = adapter.bioinfo_processed(analysis_case)

    # Don't process the case
    ready_analysis = build_analysis(
        analysis_dict=analysis_dict,
        analysis_type=analysis_type,
        valid_analysis=valid_analysis,
        current_analysis=current_analysis,
        process_case=True,
        cleanup=cleanup,
    )

    if ready_analysis:
        LOG.info("Values for %s  loaded for case %s", list(ready_analysis.keys()), analysis_case)
        LOG.info("Loaded samples in case are: %s", ", ".join(ready_analysis["samples"]))

    else:
        LOG.warning("No enteries were found for the given analysis type: %s", analysis_type)
    LOG.info("Case %s will be added/updated", analysis_case)

    load_analysis(
        adapter=adapter,
        lims_id=analysis_dict["sample"],
        processed=True,
        is_sample=False,
        dry_run=dry,
        analysis=ready_analysis,
    )



#!/usr/bin/env python

from vogue.constants.constants import MONTHS, YEARS, THIS_YEAR
from vogue.crud.find_plots.metric_per_month import value_per_month
from vogue import __version__
from vogue.settings import templates

from vogue.adapter import VogueAdapter
from vogue.settings import get_vogue_adapter
from fastapi import APIRouter, Depends, Request

router = APIRouter()


@router.get("/common/turn_around_times")
def turn_around_times(
    request: Request, year: int = THIS_YEAR, adapter: VogueAdapter = Depends(get_vogue_adapter)
):

    y_vals = [
        "received_to_delivered",
        "received_to_prepped",
        "prepped_to_sequenced",
        "sequenced_to_delivered",
    ]
    results_grouped_by_prio = {}
    results_grouped_by_cat = {}
    for y_val in y_vals:
        results_grouped_by_prio[y_val] = value_per_month(
            adapter=adapter, year=year, y_val=y_val, group_key="priority"
        )
        results_grouped_by_cat[y_val] = value_per_month(
            adapter=adapter, year=year, y_val=y_val, group_key="category"
        )

    return templates.TemplateResponse(
        "turn_around_times.html",
        context=dict(
            request=request,
            header="Turnaround Times",
            page_id="turn_around_times",
            data_prio=results_grouped_by_prio,
            data_cat=results_grouped_by_cat,
            months=[m[1] for m in MONTHS],
            version=__version__,
            year_of_interest=year,
            years=YEARS,
        ),
    )


@router.get("/common/samples")
def common_samples(
    request: Request, year: int = THIS_YEAR, adapter: VogueAdapter = Depends(get_vogue_adapter)
):
    data_cat = value_per_month(adapter=adapter, year=year, y_val="count", group_key="category")
    data_prio = value_per_month(adapter=adapter, year=year, y_val="count", group_key="priority")
    return templates.TemplateResponse(
        "samples.html",
        context=dict(
            request=request,
            header="Samples",
            page_id="samples",
            data_prio=data_prio,
            data_cat=data_cat,
            months=[m[1] for m in MONTHS],
            version=__version__,
            year_of_interest=year,
            years=YEARS,
        ),
    )



"""
    Add or update bioinfo results to bioinfo raw collection
"""
import logging
import copy
import click


from vogue.tools.cli_utils import json_read
from vogue.tools.cli_utils import dict_replace_dot
from vogue.tools.cli_utils import yaml_read
from vogue.tools.cli_utils import check_file
from vogue.tools.cli_utils import concat_dict_keys
from vogue.tools.cli_utils import add_doc as doc
from vogue.build.bioinfo_analysis import build_analysis
from vogue.load.bioinfo_analysis import load_analysis
import vogue.models.bioinfo_analysis as analysis_model

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
LOG = logging.getLogger(__name__)


@click.command(
    "raw", short_help="Read files from analysis workflows and laods in bioinfo_raw collection."
)
@click.option(
    "--sample-list",
    help="""Input list of comma separated sample names. Or specify it
within input results file under "samples" key.""",
)
@click.option(
    "-a",
    "--analysis-result",
    type=click.Path(),
    multiple=True,
    help="Input file for bioinfo analysis results. Accepted format: JSON, YAML",
)
@click.option(
    "-t",
    "--analysis-type",
    type=click.Choice(list(analysis_model.ANALYSIS_DESC.keys()) + ["all"]),
    multiple=True,
    default=["all"],
    help="Type of analysis results to load.",
)
@click.option(
    "-c",
    "--analysis-case",
    required=True,
    help="""The case that this sample belongs.
        It can be specified multiple times.""",
)
@click.option(
    "-w",
    "--analysis-workflow",
    type=click.Choice(["mip-dna", "balsamic", "microsalt"]),
    required=True,
    help="Analysis workflow used.",
)
@click.option("--workflow-version", required=True, help="Analysis workflow used.")
@click.option(
    "--case-analysis-type",
    type=click.Choice(["multiqc", "microsalt", "custom"]),
    default="multiqc",
    help="Specify the type for the case analysis. i.e. if it is multiqc output, then choose multiqc",
)
@click.option("--dry", is_flag=True, help="Load from sample or not. (dry-run)")
@doc(
    f"""
    Read and load analysis results. These are either QC or analysis output files.

    The inputs are unique ID with an analysis config file (JSON/YAML) which includes analysis results matching the
    analysis model. Analysis types recognize the following keys in the input file: {" ".join(concat_dict_keys(analysis_model.ANALYSIS_SETS,key_name=""))}
        """
)
@click.pass_context
def bioinfo_raw(
    ctx: click.Context,
    dry,
    analysis_result,
    analysis_type,
    analysis_case,
    analysis_workflow,
    workflow_version,
    case_analysis_type,
    sample_list,
):
    adapter = ctx.obj["adapter"]

    analysis_dict = dict()

    # if is_case flag is enabled, build dictionary without merging.
    # Loop over list of input config files for single sample and merge them into
    # one single dictionary

    for input_config in analysis_result:

        LOG.info("Reading and validating config file: %s", input_config)
        try:
            check_file(input_config)
        except FileNotFoundError:
            raise click.Abort()

        LOG.info("Trying JSON format")
        tmp_analysis_dict = json_read(input_config)

        if not isinstance(analysis_dict, dict):
            LOG.info("Trying YAML format")
            tmp_analysis_dict = yaml_read(input_config)
            if not isinstance(tmp_analysis_dict, dict):
                LOG.error("Cannot read input analysis config file. Type unknown.")
                raise click.Abort()

        analysis_dict = {**analysis_dict, **tmp_analysis_dict}

    analysis_dict = dict_replace_dot(analysis_dict)

    if case_analysis_type == "multiqc":
        LOG.info("--case-analysis-type set to multiqc, taking only report_saved_raw_data key")
        report_saved_raw_data = analysis_dict["report_saved_raw_data"]
        analysis_dict.clear()
        analysis_dict["report_saved_raw_data"] = copy.deepcopy(report_saved_raw_data)

    # Get current sample if any
    old_keys = list(analysis_dict.keys())
    analysis_dict[case_analysis_type] = copy.deepcopy(analysis_dict)
    for key in old_keys:
        analysis_dict.pop(key)

    valid_analysis = dict()
    analysis_dict["case"] = analysis_case
    analysis_dict["workflow"] = analysis_workflow
    analysis_dict["workflow_version"] = workflow_version
    analysis_dict["case_analysis_type"] = case_analysis_type

    if sample_list:
        sample_id = sample_list.split(",")
    else:
        if "sample" not in analysis_dict.keys():
            LOG.error("sample key not found in input json. Use --sample-list instead")
            raise click.Abort()
        # store sample_id from dict to avoid losing it downstream cleanup
        sample_id = analysis_dict["sample"]

    analysis_dict["sample"] = sample_id
    # Don't process the case
    current_analysis = adapter.bioinfo_raw(analysis_case)

    # Don't process the case
    ready_analysis = build_analysis(
        analysis_dict=analysis_dict,
        analysis_type=analysis_type,
        valid_analysis=valid_analysis,
        current_analysis=current_analysis,
        process_case=False,
        cleanup=False,
    )

    if ready_analysis:
        LOG.info("Values for %s  loaded for case %s", list(ready_analysis.keys()), analysis_case)
        LOG.info("Loaded samples in case are: %s", ", ".join(ready_analysis["samples"]))

    else:
        LOG.warning("No enteries were found for the given analysis type: %s", analysis_type)
    LOG.info("Case %s will be added/updated", analysis_case)

    load_analysis(
        adapter=adapter,
        lims_id=analysis_dict["sample"],
        processed=False,
        is_sample=False,
        dry_run=dry,
        analysis=ready_analysis,
    )



from typing import Optional

from pydantic import Field, BaseModel


class Sample(BaseModel):
    id: str = Field(..., alias="_id")
    sequenced_date: Optional[int]
    received_date: Optional[int]
    prepared_date: Optional[int]
    delivery_date: Optional[int]
    sequenced_to_delivered: Optional[int]
    prepped_to_sequenced: Optional[int]
    received_to_prepped: Optional[int]
    received_to_delivered: Optional[int]
    family: Optional[int]
    strain: Optional[str]
    source: Optional[int]
    customer: Optional[int]
    priority: Optional[int]
    initial_qc: Optional[int]
    library_qc: Optional[int]
    prep_method: Optional[str]
    sequencing_qc: Optional[int]
    application_tag: Optional[str]
    category: Optional[str]
    amount: Optional[float]
    amount_concentration: Optional[float]
    nr_defrosts: Optional[int]
    lotnr: Optional[str]
    microbial_library_concentration: Optional[float]
    library_size_pre_hyb: Optional[int]
    library_size_post_hyb: Optional[int]
    library_size_pre_hyb: Optional[int]
    library_size_post_hyb: Optional[int]

    class Config:
        allow_population_by_field_name = True



#!/usr/bin/env python

from vogue.constants.constants import LANE_UDFS


def instrument_info(adapter, year: int, metric: str) -> dict:
    """Prepares data for a plot Q30 values for diferent runs over time"""
    instruments = {"axis": {"y": "Average Q30"}, "data": {p: {} for p in LANE_UDFS}, "title": "Q30"}
    pipe = [
        {"$project": {"year": {"$year": "$date"}, "instrument": 1, "date": 1, "avg": 1}},
        {"$match": {"year": {"$eq": int(year)}}},
        {
            "$group": {
                "_id": {"instrument": "$instrument"},
                "data": {"$push": {metric: "$avg." + metric, "date": "$date", "run_id": "$_id"}},
            }
        },
    ]

    aggregate_result = adapter.flowcells_aggregate(pipe)
    for result in aggregate_result:
        group = result["_id"]["instrument"]
        for plot_name in LANE_UDFS:
            data_tuples = [(d["date"], d["run_id"], d.get(plot_name)) for d in result["data"]]
            data_sorted = sorted(data_tuples)
            data = []
            for date, run_id, value in data_sorted:
                if value:
                    data.append([date, value, run_id])
            if data:
                instruments["data"][plot_name][group] = {"data": data}
    return instruments



import logging
import copy

import vogue.models.bioinfo_analysis as analysis_model

LOG = logging.getLogger(__name__)


def inspect_analysis_result(analysis_dict: dict):
    """
    Takes input analysis_dict dictionary and validates entries.

    Checks for there are at least two keys in analysis_dict dictionary. If there is less than two, or the key doesn't
    exist, disqualifies the file and returns False
    """

    case_analysis_type = analysis_dict["case_analysis_type"]

    # detect if multiqc analysis_dict is multiqc
    if case_analysis_type == "multiqc":
        analysis_dict = analysis_dict["multiqc"]["report_saved_raw_data"]
    elif case_analysis_type == "microsalt":
        analysis_dict = analysis_dict["microsalt"]
    else:
        LOG.warning("Analysis type %s is not supported for cleanup", case_analysis_type)
        return None

    # Second level keys in ANALYSIS_SETS are bioinformatic tool name OR analysis name.
    # This line will extract all the second level keys.
    analysis_names = [
        e
        for key in analysis_model.ANALYSIS_SETS
        if isinstance(analysis_model.ANALYSIS_SETS[key], dict)
        for e in analysis_model.ANALYSIS_SETS[key].keys()
    ]

    # Now we need to know which analysis tools or bioinformatic tools in <analysis_dict>'s first level keys are among
    # valid analysis_names above. If indices needed, change i,e and extract i.
    valid_analysis = [e for e in list(analysis_dict.keys()) if e in analysis_names]

    if not valid_analysis:
        return None

    LOG.info(f"The following keys were found in the input config: {valid_analysis}")
    return valid_analysis



from statistics import mean


def reagent_category_data(
    adapter, index_category: str, flowcell_performance_treshold: float
) -> dict:
    """Preraring data for index performance view grouped on reagent label category"""

    pipe = (
        [
            {
                "$lookup": {
                    "from": "reagent_label_category",
                    "localField": "index",
                    "foreignField": "_id",
                    "as": "index_category",
                }
            },
            {"$match": {"index_category.category": index_category}},
        ]
        + _normalized_performance_pipe(flowcell_performance_treshold)
        + [
            {
                "$project": {
                    "_id": "$_id.index",
                    "avg_performance": {"$avg": "$normalized_index_performance"},
                    "nr_runs": {"$size": "$normalized_index_performance"},
                }
            }
        ]
    )

    aggregate_result = adapter.reagent_label_aggregate(pipe)

    average_normalized_peformance = []

    for data in aggregate_result:
        average_normalized_peformance.append(
            {
                "name": data["_id"],
                "y": data["avg_performance"],
                "nr_runs": data["nr_runs"],
            }
        )
    return average_normalized_peformance


def reagent_label_data(adapter, index: str, flowcell_performance_treshold: float) -> list:
    """Preraring data for index performance view grouped on reagent label"""

    pipe = [{"$match": {"index": {"$eq": index}}}] + _normalized_performance_pipe(
        flowcell_performance_treshold
    )

    aggregate_result = list(adapter.reagent_label_aggregate(pipe))

    if not aggregate_result:
        return []

    performance = aggregate_result[0]["normalized_index_performance"]
    flowcells = aggregate_result[0]["flowcell_id"]

    normalized_peformance = list(map(lambda x, y: [x, y], flowcells, performance))
    return normalized_peformance


def _normalized_performance_pipe(flowcell_performance_treshold) -> list:
    """Mongo aggregation pipe to get normalized index performance."""

    return [
        {
            "$match": {
                "flowcell_target_reads": {"$exists": "True", "$ne": None, "$gt": 0},
                "flowcell_total_reads": {"$exists": "True", "$ne": None, "$gt": 0},
                "index_target_reads": {"$exists": "True", "$ne": None, "$gt": 0},
                "index_total_reads": {"$exists": "True", "$ne": None, "$gt": 1000},
                "flowcell_id": {"$exists": "True"},
            }
        },
        {
            "$project": {
                "flowcell_id": 1,
                "index_performance": {"$divide": ["$index_total_reads", "$index_target_reads"]},
                "flowcell_performance": {
                    "$divide": ["$flowcell_total_reads", "$flowcell_target_reads"]
                },
                "index": 1,
            }
        },
        {
            "$project": {
                "normalized_index_performance": {
                    "$divide": ["$index_performance", "$flowcell_performance"]
                },
                "flowcell_id": 1,
                "flowcell_performance": 1,
                "index": 1,
            }
        },
        {"$match": {"flowcell_performance": {"$gt": flowcell_performance_treshold}}},
        {
            "$group": {
                "_id": {"index": "$index"},
                "normalized_index_performance": {"$push": "$normalized_index_performance"},
                "flowcell_performance": {"$push": "$flowcell_performance"},
                "flowcell_id": {"$push": "$flowcell_id"},
            }
        },
    ]



from fastapi import APIRouter, Depends, Request
from fastapi.responses import JSONResponse
from vogue.adapter import VogueAdapter
from vogue.crud.create import create_sample, create_bioinfo_sample
from vogue.exceptions import InsertError
from vogue.models.database import Sample, BioInfoSample
from vogue.settings import get_vogue_adapter
from vogue.tools.auth import verify_signature

router = APIRouter()


@router.post("/sample", response_model=Sample)
def sample(
    request: Request, database_sample: Sample, adapter: VogueAdapter = Depends(get_vogue_adapter)
):
    verify_signature(request=request)

    try:
        inserted_id = create_sample(adapter=adapter, sample=database_sample)
    except InsertError as e:
        return JSONResponse(content=e.message)

    return JSONResponse(content=f"Sample {inserted_id} was just inserted to the sample collection")


@router.post("/bioinfo_sample", response_model=BioInfoSample)
def bioinfo_sample(
    request: Request,
    database_sample: BioInfoSample,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):
    verify_signature(request=request)
    try:
        inserted_id = create_bioinfo_sample(adapter=adapter, sample=database_sample)
    except InsertError as e:
        return JSONResponse(content=e.message)

    return JSONResponse(
        content=f"Sample {inserted_id} was just inserted to the bioinfo_sample collection"
    )



#!/usr/bin/env python


from vogue.settings import templates

from vogue.adapter import VogueAdapter
from vogue.settings import get_vogue_adapter
from fastapi import APIRouter, Depends, Request

router = APIRouter()


from vogue.constants.constants import LANE_UDFS, YEARS, THIS_YEAR
from vogue.crud.find_plots.sequencing import instrument_info
from vogue import __version__


@router.get("/sequencing/runs")
async def runs(
    request: Request,
    metric: str = "% Bases >=Q30",
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):

    print(metric)
    selcted_metric = metric
    aggregate_result = instrument_info(adapter, year, selcted_metric)

    return templates.TemplateResponse(
        "runs.html",
        context=dict(
            request=request,
            header="Sequencing Instruments",
            page_id="runs",
            metric=selcted_metric,
            metrices=LANE_UDFS,
            results=aggregate_result,
            version=__version__,
            year_of_interest=year,
            years=YEARS,
        ),
    )



MASTER_STEPS_UDFS = {
    "received": {
        "steps": [
            "CG002 - Reception Control (Dev)",
            "CG002 - Reception Control",
            "Reception Control TWIST v1",
            "Reception Control no placement v1",
            "Reception Control (RNA) v1",
            "Reception Control (Dev) v3",
        ],
        "date_udf": "date arrived at clinical genomics",
    },
    "prepared": {
        "steps": [
            "CG002 - Aggregate QC (Library Validation) (Dev)",
            "Aggregate QC (Library Validation) (Dev) v3",
            "CG002 - Aggregate QC (Library Validation)",
            "Aggregate QC (Library Validation) TWIST v1",
            "Aggregate QC (Library Validation) (RNA) v2",
            "Aggregate QC (Library Validation)",
        ]
    },
    "delivery": {"steps": ["CG002 - Delivery", "Delivery v1"], "date_udf": "Date delivered"},
    "sequenced": {
        "steps": [
            "CG002 - Illumina Sequencing (HiSeq X)",
            "CG002 - Illumina Sequencing (Illumina SBS)",
            "AUTOMATED - NovaSeq Run",
        ],
        "date_udf": "Finish Date",
        "nova_seq": ["AUTOMATED - NovaSeq Run"],
        "q30r1_udf": "% Bases >=Q30 R1",
        "q30r2_udf": "% Bases >=Q30 R2",
    },
    "concentration_and_nr_defrosts": {
        "lot_nr_step": [
            "CG002 - End repair Size selection A-tailing and Adapter ligation (TruSeq PCR-free DNA)",
            "End repair Size selection A-tailing and Adapter ligation (TruSeq PCR-free DNA)",
        ],
        "concentration_step": [
            "CG002 - Aggregate QC (Library Validation)",
            "Aggregate QC (Library Validation)",
        ],
        "lot_nr_udf": "Lot no: TruSeq DNA PCR-Free Sample Prep Kit",
        "concentration_udf": "Concentration (nM)",
        "apptags": ["WGSPCF", "WGTPCF", "WGLPCF"],
    },
    "final_conc_and_amount_dna": {
        "amount_udf": "Amount (ng)",
        "concentration_udf": "Concentration (nM)",
        "concentration_step": [
            "CG002 - Aggregate QC (Library Validation)",
            "Aggregate QC (Library Validation)",
        ],
        "amount_step": ["CG002 - Aggregate QC (DNA)", "Aggregate QC (DNA)"],
        "apptags": ["WGSLIF", "WGTLIF", "WGLLIF"],
    },
    "microbial_library_concentration": {
        "concentration_step": "CG002 - Aggregate QC (Library Validation)",
        "concentration_udf": "Concentration (nM)",
        "apptags": "NX",
    },
    "library_size_pre_hyb": {
        "TWIST": {
            "size_step": ["pool samples TWIST v1"],
            "stage_udf": {"3999": "Size (bp)", "2176": "Average Size (bp)"},
        },
        "SureSelect": {
            "size_step": ["CG002 - Amplify Adapter-Ligated Library (SS XT)"],
            "size_udf": "Size (bp)",
            "apptags": ["EXO", "EFT", "PAN", "PAL"],
        },
    },
    "library_size_post_hyb": {
        "TWIST": {
            "size_step": ["CG002 - Sort HiSeq Samples"],
            "stage_udf": {"4005": "Size (bp)", "2182": "Average Size (bp)"},
        },
        "SureSelect": {
            "size_step": ["CG002 - Amplify Captured Libraries to Add Index Tags (SS XT)"],
            "size_udf": "Size (bp)",
            "apptags": ["EXO", "EFT", "PAN", "PAL"],
        },
    },
    "reagent_labels": {
        "steps": {
            "bcl": ["Bcl Conversion & Demultiplexing (Nova Seq)"],
            "define": ["Define Run Format and Calculate Volumes (Nova Seq)"],
            "pre_bcl": ["STANDARD Prepare for Sequencing (Nova Seq)"],
        },
        "udfs": {"reads": "# Reads", "target_reads": "Reads to sequence (M)"},
        "exclue_tags": ["RM", "EX", "PA"]
        ## EX and PA should be included later. Fetching fraction from Poos Samples Twist 1
    },
}

INSTRUMENTS = {
    "D00410": "Sigourney",
    "D00415": "Merida",
    "D00450": "Arnold",
    "ST-E00198": "Dorothy",
    "ST-E00201": "Irene",
    "ST-E00214": "Marie",
    "ST-E00266": "Rita",
    "ST-E00269": "Rosalyn",
    "A00187": "Ingrid",
    "A00689": "Barbara",
    "A00621": "Greta",
    "FS10000534": "Katherine",
}

TEST_SAMPLES = [
    "SIB603A6",
    "ADM1852A2",
    "SIB603A3",
    "SER305A1",
    "AND51A3",
    "BRA1907A5",
    "SIB552A11",
    "SIB603A9",
    "SIB257A27",
    "SIB552A15",
    "SIB611A3",
    "SIB552A16",
    "SIB256A3",
    "SIB601A5",
    "SIB553A24",
    "SIB404A15",
    "SIB404A13",
    "SIB663A2",
    "SIB701A1",
    "SIB552A18",
    "SIB257A26",
    "SER303A1",
    "SIB553A27",
    "SIB257A23",
    "SIB607A1",
    "SIB603A2",
    "SIB404A14",
    "SIB403A27",
    "SIB603A12",
    "SIB603A14",
    "SIB601A11",
    "SER305A2",
    "SIB663A4",
    "ADM1852A1",
    "SIB552A12",
    "SIB552A13",
    "SIB552A10",
    "SIB553A19",
    "SIB553A10",
    "SIB553A15",
    "SIB603A8",
    "SIB603A4",
    "SIB553A17",
    "SIB603A16",
    "SIB652A1",
    "SIB552A3",
    "SIB552A4",
    "SIB601A7",
    "SIB553A28",
    "SIB552A2",
    "SIB603A18",
    "BRA1907A1",
    "SIB404A17",
    "SIB404A16",
    "SIB553A3",
    "SIB256A1",
    "BIL1393A1",
    "SIB601A1",
    "SIB603A10",
    "SER305A5",
    "SIB603A5",
    "SER305A4",
    "ADM1851A4",
    "SER305A6",
    "SER307A8",
    "BRA1907A4",
    "BRA1907A3",
    "SIB601A4",
    "SER307A7",
    "BRA1907A2",
    "SIB257A25",
    "SIB603A13",
    "SIB601A12",
    "SIB601A14",
    "SER305A7",
    "SIB552A17",
    "SIB256A2",
    "SIB553A14",
    "SIB601A2",
    "SIB553A12",
    "SIB603A20",
    "SIB553A13",
    "SIB553A9",
    "SIB404A19",
    "ADM1852A3",
    "SIB606A2",
    "SIB404A18",
    "SIB552A5",
    "SIB603A1",
    "SIB552A6",
    "SER305A3",
    "SIB552A8",
    "SIB351A22",
    "SER307A4",
    "SIB603A15",
    "SER305A8",
    "SIB603A19",
    "SIB553A7",
    "SIB611A1",
    "SIB553A8",
    "SIB553A5",
    "SER303A3",
    "SIB351A21",
    "SIB553A16",
    "SIB255A1",
    "SIB608A1",
    "SIB351A24",
    "SIB501A2",
    "SIB403A25",
    "SIB501A3",
    "SER303A7",
    "SIB603A11",
    "AND51A4",
    "SIB553A25",
    "SIB502A2",
    "SIB501A4",
    "SER303A8",
    "SIB652A2",
    "SIB257A24",
    "AND51A1",
    "BRA1907A6",
    "AND51A5",
    "SIB501A1",
    "SIB601A8",
    "SIB257A22",
    "SIB653A1",
    "SER307A2",
    "SER303A6",
    "SIB351A20",
    "SER307A3",
    "SIB553A6",
    "ADM1852A4",
    "SIB601A6",
    "SIB404A21",
    "SER303A4",
    "SER303A5",
    "SER303A2",
    "SIB601A10",
    "SIB553A26",
    "SER307A5",
    "SIB553A1",
    "SIB253A1",
    "SIB351A23",
    "SIB253A2",
    "SIB404A22",
    "SIB601A3",
    "SIB553A4",
    "SIB702A1",
    "ADM1851A3",
    "SIB552A14",
    "AND51A6",
    "SIB609A1",
    "SER307A6",
    "SIB606A1",
    "SIB611A2",
    "SIB553A22",
    "SIB255A3",
    "SIB603A17",
    "SIB253A3",
    "ADM1851A2",
    "SIB601A13",
    "SIB553A18",
    "AND51A2",
    "SIB403A26",
    "SIB552A9",
    "SIB553A2",
    "SIB553A20",
    "SER307A1",
    "SIB553A21",
    "SIB404A20",
    "SIB663A3",
    "SIB404A24",
    "ADM990A1",
    "SIB603A7",
    "SIB404A23",
    "SIB553A23",
    "SIB552A7",
    "SIB502A1",
    "SIB351A19",
    "SIB552A1",
    "SIB601A9",
    "ADM1851A1",
    "SIB553A11",
    "SIB255A2",
]

LANE_UDFS = [
    "Clusters PF R1",
    "Clusters PF R2",
    "% Aligned R1",
    "% Aligned R2",
    "% Bases >=Q30",
    "% Bases >=Q30 R1",
    "% Bases >=Q30 R2",
    "% Error Rate R1",
    "% Error Rate R2",
    "% Phasing R1",
    "% Prephasing R1",
    "% Prephasing R2",
    "%PF R1",
    "%PF R2",
    "Cluster Density (K/mm^2) R1",
    "Cluster Density (K/mm^2) R2",
    "Intensity Cycle 1 R1",
    "Intensity Cycle 1 R2",
    "Reads PF (M) R1",
    "Reads PF (M) R2",
    "Yield PF (Gb) R1",
    "Yield PF (Gb) R2",
    "% Phasing R2",
]



import logging
import click
from vogue.load.sample import load_one, load_all, load_recent, load_one_dry, load_all_dry
from datetime import date, timedelta

from genologics.entities import Sample

LOG = logging.getLogger(__name__)


@click.command("sample", short_help="load sample/samples into db.")
@click.option("-s", "--sample-lims-id", help="Input sample lims id")
@click.option(
    "-a",
    "--all_samples",
    is_flag=True,
    help="Loads all lims samples if no other options are selected",
)
@click.option(
    "-f",
    "--load-from",
    help="Use together with --all_samples. Load from this sample lims id. Use if load all broke. Start where it ended",
)
@click.option(
    "-d", "--days", type=int, help="Update only samples updated in the latest number of days"
)
@click.option("--dry", is_flag=True, help="Load from sample or not. (dry-run)")
@click.pass_context
def sample(ctx: click.Context, sample_lims_id, all_samples, load_from, days, dry):
    """Read and load lims data for one sample, all samples or the most recently updated samples."""
    adapter = ctx.obj["adapter"]
    lims = ctx.obj["lims"]

    if not lims:
        LOG.warning("Lims connection failed.")
        raise click.Abort()

    lims = lims

    if days:
        some_days_ago = date.today() - timedelta(days=days)
        the_date = some_days_ago.strftime("%Y-%m-%dT00:00:00Z")
        load_recent(adapter, lims, the_date)
    elif all_samples:
        if dry:
            load_all_dry()
        else:
            load_all(adapter, lims=lims, start_sample=load_from)
    elif sample_lims_id:
        lims_sample = Sample(lims, id=sample_lims_id)
        if not lims_sample:
            LOG.critical("The sample does not exist in the database in the LIMS database.")
            raise SyntaxError()
        if dry:
            load_one_dry(lims_sample)
        else:
            load_one(adapter, lims_sample, lims=lims)



import logging
import click
from vogue.load.flowcell import load_one, load_all, load_recent
from datetime import date, timedelta

from vogue.constants.constants import RUN_TYPES

LOG = logging.getLogger(__name__)


@click.command("flowcell", short_help="Load flowcell into db.")
@click.option("-r", "--run-id", help="Run id for the run. Eg: 190510_A00689_0032_BHJLW2DSXX")
@click.option("-a", "--all-runs", is_flag=True, help="Loads all flowcells found in LIMS.")
@click.option("--dry", is_flag=True, help="Load from flowcell or not. (dry-run)")
@click.option(
    "-d", "--days", type=int, help="Update only runs updated in the latest number of days"
)
@click.pass_context
def flowcell(ctx: click.Context, run_id, all_runs, dry, days):
    """Read and load lims data for a one or all many runs"""
    adapter = ctx.obj["adapter"]
    lims = ctx.obj["lims"]
    if not lims:
        LOG.warning("Lims connection failed.")
        raise click.Abort()

    lims = lims

    if days:
        try:
            some_days_ago = date.today() - timedelta(days=days)
            the_date = some_days_ago.strftime("%Y-%m-%dT00:00:00Z")
        except Exception as err:
            LOG.error(err)
            raise click.Abort()
        load_recent(adapter, lims, the_date)
    elif all_runs:
        load_all(adapter, lims=lims)
    elif run_id:
        runs = []
        for run_type in RUN_TYPES:
            runs.extend(lims.get_processes(udf={"Run ID": run_id}, type=run_type))
        if runs == []:
            LOG.warning("There is no run with this Run ID")
            raise click.Abort()
        if len(runs) > 1:
            LOG.warning("There is more than one run with this run ID. Picking the latest")
            run = runs[-1]
        else:
            run = runs[0]
        load_one(adapter, run=run)



#!/usr/bin/env python


from vogue.adapter import VogueAdapter
from vogue.settings import get_vogue_adapter, templates

from fastapi import APIRouter, Depends, Request

router = APIRouter()

from vogue.constants.constants import THIS_YEAR
from vogue.crud.find_plots.reagent_labels import reagent_label_data, reagent_category_data
from vogue import __version__

flowcell_performance_treshold = 0.3


@router.get("/reagent_labels")
async def reagent_labels(
    request: Request,
    index_category: str = "Illumina IDT UD",
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):

    aggregate_result = reagent_category_data(adapter, index_category, flowcell_performance_treshold)
    return templates.TemplateResponse(
        "reagent_labels.html",
        context=dict(
            request=request,
            flowcell_performance_treshold=flowcell_performance_treshold,
            header="Overall performance per index",
            nr_indexes=len(aggregate_result),
            index_category=index_category,
            index_categories=adapter.get_reagent_label_categories(),
            year_of_interest=year,
            results=aggregate_result,
            version=__version__,
        ),
    )


@router.get("/reagent_label")
async def reagent_label(
    request: Request,
    reagent_label: str = "G09 - UDI0071,Illumina IDT UD",
    year: int = THIS_YEAR,
    adapter: VogueAdapter = Depends(get_vogue_adapter),
):
    label, index_category = reagent_label.split(",")

    aggregate_result = reagent_label_data(adapter, label, flowcell_performance_treshold)
    index_categories = list(adapter.get_all_reagent_label_names_grouped_by_category())
    return templates.TemplateResponse(
        "reagent_label.html",
        context=dict(
            request=request,
            flowcell_performance_treshold=flowcell_performance_treshold,
            header="Normalized index performance per flowcell",
            index_category=index_category,
            year_of_interest=year,
            reagent_label=label,
            index_categories=index_categories,
            results=aggregate_result,
            version=__version__,
        ),
    )



from vogue.build.flowcell import build_run
from vogue.constants.lims_constants import MASTER_STEPS_UDFS, INSTRUMENTS
import logging

LOG = logging.getLogger(__name__)

SEQUENCING_STEPS = MASTER_STEPS_UDFS["sequenced"]["steps"]


def load_one(adapter, run):
    """Function to load one lims flowcell into the database"""
    run_id = run.udf.get("Run ID")
    if not run_id:
        LOG.warning("Run ID is missing")
        return
    date, instrument = run_id.split("_")[0:2]
    instrument_name = INSTRUMENTS.get(instrument)
    if not instrument_name:
        LOG.warning("Could not get instrument name")
        return
    mongo_run = build_run(run=run, instrument=instrument_name, date=date)
    adapter.add_or_update_document(mongo_run, adapter.flowcell_collection)


def load_all(adapter, lims):
    """Function to load all lims flowcell into the database"""
    for run in lims.get_processes(type=SEQUENCING_STEPS):
        load_one(adapter, run)


def load_recent(adapter, lims, the_date):
    """Function to load all lims flowcell into the database"""
    for run in lims.get_processes(type=SEQUENCING_STEPS, last_modified=the_date):
        load_one(adapter, run)



from pathlib import Path

from genologics.lims import Lims
from pymongo import MongoClient
from pydantic import BaseSettings

from vogue.adapter import VogueAdapter

VOGUE_PACKAGE = Path(__file__).parent
PACKAGE_ROOT: Path = VOGUE_PACKAGE.parent
ENV_FILE: Path = PACKAGE_ROOT / ".env"
from fastapi.templating import Jinja2Templates

TEMPLATES_DIR: Path = VOGUE_PACKAGE / "api" / "api_v1" / "templates"
from fastapi.staticfiles import StaticFiles

STATIC_FILES: Path = VOGUE_PACKAGE / "api" / "api_v1" / "static"
static_files = StaticFiles(directory=str(STATIC_FILES))

templates = Jinja2Templates(directory=str(TEMPLATES_DIR))


class Settings(BaseSettings):
    """Settings for serving the vogue app and connect to the mongo database"""

    lims_base_uri: str = "dummy"
    lims_username: str = "dummy"
    lims_password: str = "dummy"
    db_uri: str = "mongodb://localhost:27017/vogue-demo"
    db_name: str = "vogue-demo"
    secret_key: str = "dummy"
    algorithm: str = "ABC"
    host: str = "localhost"
    access_token_expire_minutes: int = 60
    port: int = 8000
    hmac_key: str = "secret_key"

    class Config:
        env_file = str(ENV_FILE)


settings = Settings()


def get_lims() -> Lims:
    """Temporarily untill we have load from arnold in place"""
    return Lims(settings.lims_base_uri, settings.lims_username, settings.lims_password)


def get_vogue_adapter():
    client = MongoClient(settings.db_uri)
    return VogueAdapter(client, db_name=settings.db_name)



import hashlib
import hmac
import secrets

from fastapi import HTTPException, Request

from vogue.settings import settings


async def verify_signature(
    request: Request,
) -> None:
    digester = hmac.new(key=settings.hmac_key.encode(), digestmod=hashlib.sha256)
    request_body = await request.body()
    digester.update(request_body)
    received_digest = request.headers.get("authorization")
    expected_digest = "hmac " + digester.hexdigest()
    if not secrets.compare_digest(received_digest, expected_digest):
        raise HTTPException(status_code=401, detail="Invalid signature")



from vogue.constants.lims_constants import LANE_UDFS, MASTER_STEPS_UDFS
import logging
LOG = logging.getLogger(__name__)


def filter_none(mongo_dict):
    """Function to filter out Nones and NaN from a dict."""
    for key in list(mongo_dict.keys()):
        val = mongo_dict[key]
        try:
            if val is None or math.isnan(val):
                mongo_dict.pop(key)
        except Exception:
            continue
    return mongo_dict


def get_define_step_data(pool):
    """Search the artifact history for the define steps.
    Input:
        pool: lims artifact - Input to bcl step
    Returns:
        define_step_outputs: dict
            keys: sample ids
            value: target reads (udf 'Reads to sequence (M)')
        define_step: lims process
            the define step
        flowcell_target_reads: int
            the summe of the udf 'Reads to sequence (M)' from all outarts in the step
    """

    LOG.info('Searching Define Steps')

    process = pool.parent_process

    while process.type.name not in MASTER_STEPS_UDFS['reagent_labels'][
            'steps']['define']:
        art = process.all_inputs()[0]
        process = art.parent_process

    define_step_outputs = {}
    flowcell_target_reads = 0

    for art in process.all_outputs():
        if art.type != 'Analyte':
            continue

        index_target_reads = _get_target_reads(art)
        if index_target_reads:
            flowcell_target_reads += index_target_reads

        if len(art.samples) != 1:  # ignore pools
            continue
        sample_id = art.samples[0].id
        define_step_outputs[sample_id] = index_target_reads

    LOG.info('Done')

    return define_step_outputs, flowcell_target_reads, process


def _get_target_reads(artifact):

    udf = MASTER_STEPS_UDFS['reagent_labels']['udfs']['target_reads']
    index_target_reads = artifact.udf.get(udf)
    if index_target_reads and index_target_reads.isnumeric():
        return int(index_target_reads) * 1000000
    else:
        return None


def reagent_label_data(bcl_step):
    """This function takes as input a bcl conversion and demultiplexing step. From that step it goes
    back in artifact history to the prevoius Define step. Both step types exist in the Nova Seq 
    workflow. 
    From the output artifacts of the bcl step, index_total_reads is calculated:
        index_total_reads: the sum of '# Reads' from all artifact with a specific index
        flowcell_total_reads: the sum of '# Reads' from all output artifacts
    From the output artifacts of the define step, index_target_reads and flowcell_target_reads are fetched:
        index_target_reads: fetched from the 'Reads to sequence (M)' udf of the output artifact with a specific index 
        flowcell_target_reads: the sum of the 'Reads to sequence (M)' udf of all the output artifacts
        """

    define_step = None
    indexes = {}
    flowcell_total_reads = 0

    for inp, outp in bcl_step.input_output_maps:
        pre_bcl_steps = MASTER_STEPS_UDFS['reagent_labels']['steps']['pre_bcl']
        if inp['parent-process'].type.name not in pre_bcl_steps:
            continue

        if outp['output-generation-type'] != 'PerReagentLabel':
            continue

        lane = inp['uri']
        art = outp['uri']

        index_reads = art.udf.get(
            MASTER_STEPS_UDFS['reagent_labels']['udfs']['reads'])

        if index_reads is None or art.qc_flag == 'FAILED':
            continue

        flowcell_total_reads += index_reads

        sample = art.samples[0]  # Will always be only one sample in the list
        application_tag = sample.udf.get('Sequencing Analysis')

        if application_tag[0:2] in MASTER_STEPS_UDFS['reagent_labels'][
                'exclue_tags']:
            continue

        if not define_step:
            define_step_outputs, flowcell_target_reads, define_step = get_define_step_data(
                lane)

        if not sample.id in define_step_outputs:
            LOG.info('This sample whas put as a pool into the define step: ' +
                     sample.id + ' ' + application_tag)
            continue

        index = art.reagent_labels[0]

        container, lane_nr = lane.location
        if index not in indexes:
            indexes[index] = {
                '_id': '_'.join([index, container.name]),
                'url': index.replace(' ', ''),
                'index_total_reads': index_reads,
                'index_target_reads': define_step_outputs[sample.id],
                'flowcell_target_reads': flowcell_target_reads,
                'index': index,
                'sample': sample.id,
                'lanes': {
                    lane_nr: dict(art.udf.items())
                },
                'flowcell_id': container.name,
                'flowcell_type': container.type.name,
                'define_step_udfs': dict(define_step.udf.items()),
                'define_step_id': define_step.id,
                'bcl_step_id': bcl_step.id
            }
        else:
            indexes[index]['lanes'][lane_nr] = dict(art.udf.items())
            indexes[index]['index_total_reads'] += index_reads

    for index, data in indexes.items():
        data['flowcell_total_reads'] = flowcell_total_reads
        indexes[index] = filter_none(data)

    return indexes
